/* 
 * Copyright (c) 2004 Nicolas Rougier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this software; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <sstream>
#include <SDL/SDL_opengl.h>
#include <SDL/SDL.h>
#include <SDL/SDL_image.h>
#include "timer.h"
#include "log.h"
#include "screen.h"
#include "camera.h"
#include "texture.h"
#include "fx_effect.h"


void key_press (SDL_keysym *keysym, Screen &screen, Timer &timer)
{
    switch (keysym->sym) {
	case SDLK_ESCAPE:
		Mix_CloseAudio();
		SDL_Quit ();
		exit (EXIT_SUCCESS);
	    break;
	case SDLK_s:
		screen.screenshot ();
		break;
    case SDLK_p:
	case SDLK_SPACE:
        if (timer.paused())
            timer.unpause();
        else
            timer.pause();
		break;
	case SDLK_F1:
		screen.full();
	    break;
	default:
	    break;
	}
}

int main (int argc, char **argv)
{
	// Initialize random generator
	srand (1);

    std::string path;

	if (argc >= 2)
        path = argv[1];
    else
        path = DATADIR;
    
    if (path.empty())
        path = DATADIR;
    if (path[path.size()-1] != '/')
        path += "/";



	// Logo (sort of)
    //	std::string s = path + "png/boom.png";
	std::string s = path + "copyright.png";
	SDL_Surface *copyright = IMG_Load (s.c_str());

	// Check if user provided the graphic directory
	if (copyright == 0) {
        log_error ("Cannot load %s.\n", s.c_str());
		log_info ("Usage : boom <data directory>\n");
		exit (EXIT_FAILURE);
	}


	SDL_Rect src;
	SDL_Rect dst;

	// Start SDL with audio support
	if (SDL_Init (SDL_INIT_AUDIO) == -1) {
		log_error ("SDL_Init: %s", SDL_GetError ());
		exit (EXIT_FAILURE);
	}
	// Open 44.1KHz, signed 16bit, system byte order,
	//      stereo audio, using 1024 byte chunks
	if (Mix_OpenAudio (44100, MIX_DEFAULT_FORMAT, 2, 1024) == -1) {
		log_error ("Mix_OpenAudio: %s", Mix_GetError());
		exit (EXIT_FAILURE);
	}

	Screen screen (800,600);
	Camera camera (&screen);
	Timer timer;
    TextureManager::it()->init();





	// Event loop
    int done = false;
    SDL_Event event;
    int active = true;
    float time = 0;

	// Explosion
	fx::Explosion effect (path);
	//fx::Burn effect(path);
	effect.init();

    SDL_WM_SetCaption ("Boom demo", NULL);


    while (!done) {
	    while (SDL_PollEvent(&event)) {
		    switch (event.type) {
			case SDL_ACTIVEEVENT:
				if (event.active.state & SDL_APPACTIVE) {
					if (event.active.gain == 0) {
						active = false;
						timer.pause();
					}
					else {
						active = true;
						timer.unpause();
					}
				}
			    break;			    
			case SDL_VIDEORESIZE:
				screen.resize (event.resize.w, event.resize.h);
			    break;
			case SDL_KEYDOWN:
			    key_press (&event.key.keysym, screen, timer);
			    break;
			case SDL_QUIT:
			    done = 1;
			    break;
			default:
			    break;
			}
		}
	    if (active) {
			camera.look();
			effect.render ();

            src.x = 0;
            src.y = 0;
            src.w = copyright->w;
            src.h = copyright->h;
            dst.x = (screen.width()-copyright->w)/2;
            dst.y = screen.height() - copyright->h - 5;
            dst.w = copyright->w;
            dst.h = copyright->h;

            // Copyright
            if (time < 3.0f)
                screen.blit_surface (copyright, src, dst, 1);
            else if (time < 4.0f)
                screen.blit_surface (copyright, src, dst, 4.0f-time);
            


			float dt = timer.wait_next_tick();
            time += dt;

			effect.step_in_time (dt);
			screen.update();
		}
	}
	
	Mix_CloseAudio();
    SDL_Quit ();
    exit (EXIT_SUCCESS);
	return 0;
}

