/* 
 * Copyright (c) 2004 Thomas Jahn & Nicolas Rougier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this software; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

// ================================================================================
// Notes
// --------------------------------------------------------------------------------
// This code is adapted from "PixelprachtFX Demo" by Thomas Jahn aka Lithander.
// For more stuff, please visit website at www.pixelpracht.net
// ================================================================================

#include <cmath>
#include "log.h"
#include "fx_cloud.h"

using namespace fx;


// ================================================================================
// 
// --------------------------------------------------------------------------------
// 
// ================================================================================
Cloud::Cloud (Effect *effect, unsigned int n)
{
	effect_ = effect;
	particles_.resize (n);
	for (unsigned i=0; i<n; i++) {
		particles_[i].dead()  = true;
		particles_[i].cloud() = this;
	}
	first_dead_ = 0;

	type_           = FX_POINT;
	texture_        = 0;
	tile_.min_u = tile_.min_v = 0.0f;
	tile_.max_u = tile_.max_v = 1.0f;
	elongation_     = 2.0f;
	blend_          = FX_OPAQUE;
	fade_in_stop_   = 0.0f;
	fade_out_start_ = 1.0f;
	zsort_          = true;
}


// ================================================================================
// 
// --------------------------------------------------------------------------------
// 
// ================================================================================
Cloud::~Cloud (void)
{
}

// ================================================================================
// 
// --------------------------------------------------------------------------------
// 
// ================================================================================
void Cloud::add (Particle &p, unsigned int n)
{
	while ((n) &&  (first_dead_ < int(particles_.size()))) {
		particles_[first_dead_] = p;
		if (zsort_)
			effect_->add (&particles_[first_dead_]);
		first_dead_++;
		n--;
	}
}

// ================================================================================
// 
// --------------------------------------------------------------------------------
// 
// ================================================================================
void Cloud::kill (int index)
{
	// bad index
	if (index > first_dead_)
		return;

	// kill'em all
	if (index == -1) {
		for (unsigned int i=0; i<particles_.size(); i++) {
			particles_[i].dead() = true;
			effect_->remove (&particles_[i]);
		}
		first_dead_ = 0;
		return;
	}

	// already dead
	if (particles_[index].dead() == true)
		return;

	// kill it
	particles_[index] = particles_[first_dead_-1];
	particles_[first_dead_-1].dead() = true;
	effect_->remove (&particles_[first_dead_-1]);
	first_dead_--;
}

// ================================================================================
// 
// --------------------------------------------------------------------------------
// 
// ================================================================================
void Cloud::step_in_time (const float dt)
{
	int i = 0;
	while (i < first_dead_) {
		particles_[i].position().x += particles_[i].velocity().x*dt;
		particles_[i].position().y += particles_[i].velocity().y*dt;
		particles_[i].position().z += particles_[i].velocity().z*dt;
		particles_[i].rotation()    = fmod (particles_[i].rotation()-particles_[i].spin()*dt, 360.0f);
		particles_[i].age()        += dt;
		if (particles_[i].age() >= particles_[i].lifespan())
			kill (i);
		else
			i++;
	};
}


// ================================================================================
// manipulation
// --------------------------------------------------------------------------------
// 
// ================================================================================
void Cloud::scale (const unsigned mask, const vector &v)
{
	for (int i=0; i < first_dead_; i++)
		particles_[i].scale (mask, v);
}
void Cloud::scale (const unsigned mask, const float scale)
{
	for (int i=0; i < first_dead_; i++)
		particles_[i].scale (mask, scale);
}
void Cloud::scale (const unsigned mask, const float scale_x, const float scale_y, const float scale_z)
{
	for (int i=0; i < first_dead_; i++)
		particles_[i].scale (mask, scale_x, scale_y, scale_z);
}


// ================================================================================
// physics
// --------------------------------------------------------------------------------
// 
// ================================================================================
void Cloud::bullet (const vector &target)
{
	for (int i=0; i < first_dead_; i++)
		particles_[i].bullet (target);
}
void Cloud::bullet (const float tx, const float ty, const float tz)
{
	for (int i=0; i < first_dead_; i++)
		particles_[i].bullet (tx, ty, tz);
}
void Cloud::linear_gravity (const vector &point, const float gravity, const float dt)
{
	for (int i=0; i < first_dead_; i++)
		particles_[i].linear_gravity (point, gravity, dt);
}
void Cloud::linear_gravity (const float px, const float py, const float pz, const float gravity, const float dt)
{
	for (int i=0; i < first_dead_; i++)
		particles_[i].linear_gravity (px, py, pz, gravity, dt);
}
void Cloud::circular_gravity (const vector &point, const float gravity, const float dt)
{
	for (int i=0; i < first_dead_; i++)
		particles_[i].circular_gravity (point, gravity, dt);
}
void Cloud::circular_gravity (const float px, const float py, const float pz, const float gravity, const float dt)
{
	for (int i=0; i < first_dead_; i++)
		particles_[i].circular_gravity (px, py, pz, gravity, dt);
}
void Cloud::point_gravity (const vector &point, const float radius, const float mass, const float dt)
{
	for (int i=0; i < first_dead_; i++)
		particles_[i].point_gravity (point, radius, mass, dt);
}
void Cloud::point_gravity (const float px, const float py, const float pz, const float radius, const float mass, const float dt)
{
	for (int i=0; i < first_dead_; i++)
		particles_[i].point_gravity (px, py, pz, radius, mass, dt);
}
void Cloud::friction (const float viscosity, const float dt)
{
	for (int i=0; i < first_dead_; i++)
		particles_[i].friction (viscosity, dt);
}
void Cloud::flow (const vector &direction, const float viscosity, const float dt)
{
	for (int i=0; i < first_dead_; i++)
		particles_[i].flow (direction, viscosity, dt);
}
void Cloud::flow (const float dx, const float dy, const float dz, const float viscosity, const float dt)
{
	for (int i=0; i < first_dead_; i++)
		particles_[i].flow (dx, dy, dz, viscosity, dt);
}


// ================================================================================
// render 
// --------------------------------------------------------------------------------
// 
// ================================================================================
void Cloud::render (void)
{
	// texture
	if ((type_ == FX_TEX_QUAD) || (type_ == FX_TEX_SPARK) || (type_ == FX_SPRITE)) {
		glEnable (GL_TEXTURE_2D);
		glBindTexture (GL_TEXTURE_2D, texture_->get_id());
	} else {
		glDisable (GL_TEXTURE_2D);
	}

	// blend mode
	if (type_ == FX_SPRITE) {
		if (blend_ == FX_OPAQUE) {
			glEnable (GL_BLEND);
			glBlendFunc (GL_ONE, GL_ZERO);
		} else if (blend_ == FX_GLOW) {
			glEnable (GL_BLEND);
			glBlendFunc (GL_SRC_ALPHA, GL_ONE);
		} else if (blend_ == FX_TRANSPARENT) {
			glEnable (GL_BLEND);
			glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		} else if (blend_ == FX_LIGHT) {
			glEnable (GL_BLEND);
			glBlendFunc (GL_SRC_ALPHA, GL_ONE);
		}
	} else {
		if (blend_ == FX_OPAQUE) {
			glEnable (GL_BLEND);
			glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		} else if (blend_ == FX_GLOW) {
			glEnable(GL_BLEND);
			glBlendFunc(GL_ONE_MINUS_DST_COLOR, GL_ONE);
		} else if (blend_ == FX_TRANSPARENT) {
			glEnable(GL_BLEND);
			glBlendFunc(GL_SRC_COLOR, GL_ONE);
		} else if (blend_ == FX_LIGHT) {
			glEnable(GL_BLEND);
			glBlendFunc(GL_ONE, GL_ONE);
		}
	}
}

