/* 
 * Copyright (c) 2004 Thomas Jahn & Nicolas Rougier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this software; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

// ================================================================================
// Notes
// --------------------------------------------------------------------------------
// This code is adapted from "PixelprachtFX Demo" by Thomas Jahn aka Lithander.
// For more stuff, please visit website at www.pixelpracht.net
// ================================================================================

#include "log.h"
#include "texture.h"
#include "fx_effect.h"

using namespace fx;


// ================================================================================
//  
// --------------------------------------------------------------------------------
//  
// ================================================================================
void Explosion::init (void)
{
	// Get some sound
	std::string s = path_;
	s += "explosion.wav";
	sample_ = Mix_LoadWAV (s.c_str());
	if (!sample_) {
		log_warning ("Mix_LoadWAV: %s", Mix_GetError());
	}

	TextureManager *tmanager = TextureManager::it();
	last_emission_ = 0;
	emission_rate_ = .1;

	Cloud *c = new Cloud (this, 1000);
	c->type()           = FX_TEX_QUAD;
	c->blend()          = FX_TRANSPARENT;
	c->texture()        = tmanager->get (path_+"smoke.jpg");
	c->fade_in_stop()   = 0.15f;
	c->fade_out_start() = 0.95f;
	clouds_["Smoke"] 	= c;

	c = new Cloud (this, 500);
	c->type()           = FX_SPRITE;
	c->blend()          = FX_TRANSPARENT;
	c->texture()        = tmanager->get (path_+"junk.png", true);
	c->fade_in_stop()   = 0.5f;
	c->fade_out_start() = 1.0f;
	clouds_["Junk"] 	= c;

	c = new Cloud (this, 1000);
	c->type()           = FX_TEX_QUAD;
	c->blend()          = FX_LIGHT;
	c->texture()        = tmanager->get (path_+"glow.jpg");
	c->fade_in_stop()   = 0.45f;
	c->fade_out_start() = 0.45f;
	clouds_["Fire"]  = c;

	c = new Cloud (this, 1000);
	c->type()           = FX_SPRITE;
	c->blend()          = FX_TRANSPARENT;
	c->texture()        = tmanager->get (path_+"dark-smoke.png", true);
	c->fade_in_stop()   = 0.3f;
	c->fade_out_start() = 0.6f;
	clouds_["Flame"]  = c;

	templates_["SmokeMin"].lifespan() = 2.0f;
	templates_["SmokeMin"].size()     = 0.5f;
	templates_["SmokeMin"].sat()      = 0.5f;
	templates_["SmokeMin"].density()  = 0.5f;
	templates_["SmokeMin"].color()    = vector (0.6f, 0.6f, 0.6f);

	templates_["SmokeMax"].lifespan() = 2.5f;
	templates_["SmokeMax"].size()     = 0.7f;
	templates_["SmokeMax"].sat()      = 0.7f;
	templates_["SmokeMax"].density()  = 1.0f;
	templates_["SmokeMax"].color()    = vector (0.7f, 0.6f, 0.6f);

	templates_["JunkMin"].lifespan() = 2.0f;
	templates_["JunkMin"].size()     = 0.02f;
	templates_["JunkMin"].sat()      = 0.9f;
	templates_["JunkMin"].density()  = 5.0f;
	templates_["JunkMin"].color()    = vector (1.0f, 0.7f, 0.7f);

	templates_["JunkMax"].lifespan() = 3.0f;
	templates_["JunkMax"].size()     = 0.05f;
	templates_["JunkMax"].sat()      = 1.0f;
	templates_["JunkMax"].density()  = 10.0f;
	templates_["JunkMax"].color()    = vector (1.0f, 0.8f, 0.7f);

	templates_["FireMin"].lifespan() = 1.0f;
	templates_["FireMin"].size()     = 0.2f;
	templates_["FireMin"].sat()      = 0.4f;
	templates_["FireMin"].density()  = 0.4f;
	templates_["FireMin"].color()    = vector (1.0f, 1.0f, 1.0f);

	templates_["FireMax"].lifespan() = 2.0f;
	templates_["FireMax"].size()     = 0.3f;
	templates_["FireMax"].sat()      = 0.9f;
	templates_["FireMax"].density()  = 0.6f;
	templates_["FireMax"].color()    = vector (1.0f, 1.0f, 1.0f);

	templates_["FlameMin"].lifespan() = 1.0f;
	templates_["FlameMin"].size()     = 0.3f;
	templates_["FlameMin"].sat()      = 0.3f;
	templates_["FlameMin"].spin()     = -20.0f;
	templates_["FlameMin"].density()  = 0.1f;
	templates_["FlameMin"].color()    = vector (1.0f, 1.0f, 1.0f);

	templates_["FlameMax"].lifespan() = 1.0f;
	templates_["FlameMax"].size()     = 0.4f;
	templates_["FlameMax"].sat()      = 0.5f;
	templates_["FlameMax"].spin()     = 20.0f;
	templates_["FlameMax"].density()  = 0.4f;
	templates_["FlameMax"].color()    = vector (1.0f, 1.0f, 1.0f);
}


// ================================================================================
//  
// --------------------------------------------------------------------------------
//  
// ================================================================================
void Explosion::step_in_time (const float dt) {
	for (std::map<std::string, Cloud *>::iterator i=clouds_.begin(); i != clouds_.end(); i++)
		i->second->step_in_time (dt);

	// Start explosion sound
	if ((sample_) && (age_ == 0)) {
		channel_ = Mix_PlayChannelTimed (-1, sample_, -1 , 4000);
	}

	age_ += dt;
	last_emission_ += dt;

	// emit particles
	if (last_emission_ >= emission_rate_) {
		last_emission_ -= emission_rate_;

		// smoke
		if ((age_ > 0.0f) && (age_ < 2.5f)) {
            Particle p(templates_["SmokeMin"], templates_["SmokeMax"]);
            p.fit_sphere (0.0f, 0.0f, 0.0f, 0.5f);
            p.scale (SIZE, ((age_/1.7f)+0.3f));
            p.scale (POS_Y, 0.7f);
            clouds_["Smoke"]->add (p);
		}

/*		
		// junk
		if ((age_ < 1.0f) && (age_ > 0.5f)) {
			for (int i = 0; i<4; i++) {
				Particle p(templates_["JunkMin"], templates_["JunkMax"]);
				p.fit_sphere (0.0f, 0.0f, 0.0f, 0.3f);
				clouds_["Junk"]->add (p);
			}
		}
*/

		// flame & fire
		if (age_ < 1.5f) {
			for (int i = 0; i< int(2+age_); i++) {
				if (age_ > 0.5f) {
					for (int j=0; j<int(2+age_); j++) {
						Particle p(templates_["FlameMin"], templates_["FlameMax"]);
						p.fit_sphere (0.0f, 0.0f, 0.0f, age_/3+0.03f);
						p.scale (SIZE, ((age_/1.2f)+0.4f));
						p.scale (POS_Y, 0.7f);
						clouds_["Flame"]->add (p);
					}
				}
				if (age_ < 1.0f) {
					for (int j=0; j< int(2+age_); j++) {
						Particle p(templates_["FireMin"], templates_["FireMax"]);
						p.fit_sphere (0.0f, 0.0f, 0.0f, age_/2.5f);
						p.scale (SIZE, age_/0.7f+0.2f);
						p.scale (POS_Y, 0.7f);
						clouds_["Fire"]->add(p);
					}
				}
			}
		}
	}


	// Apply Actions
	clouds_["Smoke"]->scale (SIZE, 1+dt*0.25f);
	clouds_["Smoke"]->circular_gravity (0.0f, 0.0f, 0.0f, -3.0f, dt);
    clouds_["Smoke"]->flow (0.0f, 0.2f, 0.0f, 0.3f, dt);

	clouds_["Fire"]->circular_gravity (0.0f, 0.0f, 0.0f, -3.0f, dt);
	clouds_["Fire"]->friction (0.3f, dt);

	clouds_["Flame"]->scale (SIZE,  1+dt*0.2f);
	clouds_["Flame"]->scale (COLOR, 1-dt*0.4f);
	clouds_["Flame"]->point_gravity (0.0f, 0.0f, 0.0f, 0.35f, -1.4000f, dt);
	clouds_["Flame"]->flow (0.0f, 0.0f, 0.0f, 0.3f, dt);



    /* Other version
	clouds_["Smoke"]->scale (SIZE, 1+dt*0.25f);
	clouds_["Smoke"]->circular_gravity (0.0f, 0.0f, 0.0f, -3.0f, dt*1000);
	clouds_["Smoke"]->flow (0.0f, 0.0f, 0.0f, 0.1f, dt*1000);


	clouds_["Fire"]->circular_gravity (0.0f, 0.0f, 0.0f, -3.0f, dt*1000);
	clouds_["Fire"]->friction (0.3f, dt*1000);

	clouds_["Flame"]->scale (SIZE,  1+dt*0.2f);
	clouds_["Flame"]->scale (COLOR, 1-dt*0.4f);
	clouds_["Flame"]->point_gravity (0.0f, 0.0f, 0.0f, 0.35f, -1400.0f, dt*1000);
	clouds_["Flame"]->flow (0.0f, 0.0f, 0.0f, 0.3f, dt*1000);
    */


	if (age_ > 5.0f)
		reset ();
}
