/* 
 * Copyright (c) 2004 Thomas Jahn & Nicolas Rougier 
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this software; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

// ================================================================================
// Notes
// --------------------------------------------------------------------------------
// This code is adapted from "PixelprachtFX Demo" by Thomas Jahn aka Lithander.
// For more stuff, please visit website at www.pixelpracht.net
// ================================================================================

#include <cstdlib>
#include <cmath>
#include <SDL/SDL_opengl.h>
#include "log.h"
#include "fx_cloud.h"
#include "fx_particle.h"

#ifndef M_PI
# define M_PI		3.14159265358979323846	/* pi */
# define M_PI_2		1.57079632679489661923	/* pi/2 */
# define M_PI_4		0.78539816339744830962	/* pi/4 */
#endif

using namespace fx;


// ================================================================================
//  
// --------------------------------------------------------------------------------
//  
// ================================================================================
Particle::Particle (void)
{
	position_.x = position_.y = position_.z = 0.0f;
	velocity_.x = velocity_.y = velocity_.z = 0.0f;
	density_ = 1.0f;
	mass_    = 1.0f;
	size_    = 0.1f;
	spin_    = 0.0f;
	rotation_= 0.0f;
	color_.r = color_.g = color_.b = 1.0f;
	sat_     = 1;
	lifespan_= 1.0f;
	age_     = 0;
	dead_    = false;
}

// ================================================================================
//  
// --------------------------------------------------------------------------------
//  
// ================================================================================
Particle::Particle (const Particle &p1, const Particle &p2)
{
	position_.x = p1.position_.x + (p2.position_.x - p1.position_.x) * rand ()/float(RAND_MAX);
	position_.y = p1.position_.y + (p2.position_.y - p1.position_.y) * rand ()/float(RAND_MAX);
	position_.z = p1.position_.z + (p2.position_.z - p1.position_.z) * rand ()/float(RAND_MAX);
	velocity_.x = p1.velocity_.x + (p2.velocity_.x - p1.velocity_.x) * rand ()/float(RAND_MAX);
	velocity_.y = p1.velocity_.y + (p2.velocity_.y - p1.velocity_.y) * rand ()/float(RAND_MAX);
	velocity_.z = p1.velocity_.z + (p2.velocity_.z - p1.velocity_.z) * rand ()/float(RAND_MAX);
	color_.r    = p1.color_.r    + (p2.color_.r    - p1.color_.r)    * rand ()/float(RAND_MAX);
	color_.g    = p1.color_.g    + (p2.color_.g    - p1.color_.g)    * rand ()/float(RAND_MAX);
	color_.b    = p1.color_.b    + (p2.color_.b    - p1.color_.b)    * rand ()/float(RAND_MAX);
	density_    = p1.density_    + (p2.density_    - p1.density_)    * rand ()/float(RAND_MAX);
	mass_       = p1.mass_       + (p2.mass_       - p1.mass_)       * rand ()/float(RAND_MAX);
	size_       = p1.size_       + (p2.size_       - p1.size_)       * rand ()/float(RAND_MAX);
	rotation_   = p1.rotation_   + (p2.rotation_   - p1.rotation_)   * rand ()/float(RAND_MAX);
	sat_        = p1.sat_        + (p2.sat_        - p1.sat_)        * rand ()/float(RAND_MAX);
	lifespan_   = p1.lifespan_   + (p2.lifespan_   - p1.lifespan_)   * rand ()/float(RAND_MAX);
	age_        = 0;
	dead_       = false;
}


// ================================================================================
//  
// --------------------------------------------------------------------------------
//  
// ================================================================================
Particle::Particle (const Particle &p1, const Particle &p2, const float factor)
{
	interpolate (ALL, p1, p2, factor);
	age_        = 0;
	dead_       = false;
}


// ================================================================================
//  
// --------------------------------------------------------------------------------
//  
// ================================================================================
Particle &Particle::operator = (const Particle &other)
{
	if (this == &other)
		return *this;
	position_ = other.position_;
	velocity_ = other.velocity_;
	color_    = other.color_;
	density_  = other.density_;
	mass_     = other.mass_;
	size_     = other.size_;
	rotation_ = other.rotation_;
	sat_      = other.sat_;
	age_      = other.age_;
	lifespan_ = other.lifespan_;
	dead_     = other.dead_;
	return *this;
}


// ================================================================================
//  
// --------------------------------------------------------------------------------
//  
// ================================================================================
void Particle::scale (const unsigned mask, const vector &v) 
{
	scale (mask, v.x, v.y, v.z);
}
void Particle::scale (const unsigned mask, const float s) 
{
	scale (mask, s, s, s);
}
void Particle::scale (const unsigned mask, const float s_x, const float s_y, const float s_z)
{
	if (mask & POSITION) {
		position_.x *= s_x;
		position_.y *= s_y;
		position_.z *= s_z;
	} else {
		if (mask & POS_X)
			position_.x *= s_x;
		if (mask & POS_Y)
			position_.y *= s_y;
		if (mask & POS_Z)
			position_.z *= s_z;
	}

	if (mask & VELOCITY) {
		velocity_.x *= s_x;
		velocity_.y *= s_y;
		velocity_.z *= s_z;
	} else {
		if (mask & VEL_X)
			velocity_.x *= s_x;
		if (mask & VEL_Y)
			velocity_.y *= s_y;
		if (mask & VEL_Z)
			velocity_.z *= s_z;
	}

    if (mask & COLOR) {
		color_.r *= s_x;
		color_.r = (color_.r > 1.0f) ? 1.0f : color_.r;
		color_.r = (color_.r < 0.0f) ? 0.0f : color_.r;
		color_.g *= s_y;
		color_.g = (color_.g > 1.0f) ? 1.0f : color_.g;
		color_.g = (color_.g < 0.0f) ? 0.0f : color_.g;
		color_.b *= s_z;
		color_.b = (color_.b > 1.0f) ? 1.0f : color_.b;
		color_.b = (color_.b < 0.0f) ? 0.0f : color_.b;
	};

    if (mask & MASS)
		mass_ *= s_x;
    if (mask & SIZE)
		size_ *= s_x;
    if (mask & SPIN)
		spin_ *= s_x;
    if (mask & ROTATION)
		rotation_ *= s_x;
    if (mask & DENSITY)
		density_ *= s_x;
    if (mask & SAT)
		sat_ *= s_x;
}


// ================================================================================
// addition / substraction
// --------------------------------------------------------------------------------
//  
// ================================================================================
void Particle::add (const unsigned mask, const Particle &other)
{
	add (mask, other, 1);
}
void Particle::sub (const unsigned mask, const Particle &other)
{
	add (mask, other, -1);
}
void Particle::add (const unsigned mask, const Particle &other, int sign)
{
	if (mask & POSITION) {
		position_.x += sign*other.position_.x;
		position_.y += sign*other.position_.y;
		position_.z += sign*other.position_.z;
	} else {
		if (mask & POS_X)
			position_.x += sign*other.position_.x;
		if (mask & POS_Y)
			position_.y += sign*other.position_.y;
		if (mask & POS_Z)
			position_.z += sign*other.position_.z;
	}

	if (mask & VELOCITY) {
		velocity_.x += sign*other.velocity_.x;
		velocity_.y += sign*other.velocity_.y;
		velocity_.z += sign*other.velocity_.z;
	} else {
		if (mask & VEL_X)
			velocity_.x += sign*other.velocity_.x;
		if (mask & VEL_Y)
			velocity_.y += sign*other.velocity_.y;
		if (mask & VEL_Z)
			velocity_.z += sign*other.velocity_.z;
	}

    if (mask & COLOR) {
		color_.r += sign*other.color_.r;
		color_.r = (color_.r > 1.0f) ? 1.0f : color_.r;
		color_.r = (color_.r < 0.0f) ? 0.0f : color_.r;
		color_.g += sign*other.color_.g;
		color_.g = (color_.g > 1.0f) ? 1.0f : color_.g;
		color_.g = (color_.g < 0.0f) ? 0.0f : color_.g;
		color_.b += sign*other.color_.b;
		color_.b = (color_.b > 1.0f) ? 1.0f : color_.b;
		color_.b = (color_.b < 0.0f) ? 0.0f : color_.b;
	};

    if (mask & MASS)
		mass_ += sign*other.mass_;
    if (mask & SIZE)
		size_ += sign*other.size_;
    if (mask & SPIN)
		spin_ += sign*other.spin_;
    if (mask & ROTATION)
		rotation_ += sign*other.rotation_;
    if (mask & DENSITY)
		density_ += sign*other.density_;
    if (mask & SAT)
		sat_ += sign*other.sat_;
}


// ================================================================================
// Affectation
// --------------------------------------------------------------------------------
//  
// ================================================================================
void Particle::set (const unsigned mask, const Particle &other)
{
	if (mask & POSITION) {
		position_.x = other.position_.x;
		position_.y = other.position_.y;
		position_.z = other.position_.z;
	} else {
		if (mask & POS_X)
			position_.x = other.position_.x;
		if (mask & POS_Y)
			position_.y = other.position_.y;
		if (mask & POS_Z)
			position_.z = other.position_.z;
	}

	if (mask & VELOCITY) {
		velocity_.x = other.velocity_.x;
		velocity_.y = other.velocity_.y;
		velocity_.z = other.velocity_.z;
	} else {
		if (mask & VEL_X)
			velocity_.x = other.velocity_.x;
		if (mask & VEL_Y)
			velocity_.y = other.velocity_.y;
		if (mask & VEL_Z)
			velocity_.z = other.velocity_.z;
	}

    if (mask & COLOR) {
		color_.r = other.color_.r;
		color_.g = other.color_.g;
		color_.b = other.color_.b;
	};

    if (mask & MASS)
		mass_ = other.mass_;
    if (mask & SIZE)
		size_ = other.size_;
    if (mask & SPIN)
		spin_ = other.spin_;
    if (mask & ROTATION)
		rotation_ = other.rotation_;
    if (mask & DENSITY)
		density_ = other.density_;
    if (mask & SAT)
		sat_ = other.sat_;
}


// ================================================================================
// Negation
// --------------------------------------------------------------------------------
//  
// ================================================================================
void Particle::neg (const unsigned mask)
{
	if (mask & POSITION) {
		position_.x = -position_.x;
		position_.y = -position_.y;
		position_.z = -position_.z;
	} else {
		if (mask & POS_X)
			position_.x = -position_.x;
		if (mask & POS_Y)
			position_.y = -position_.y;
		if (mask & POS_Z)
			position_.z = -position_.z;
	}

	if (mask & VELOCITY) {
		velocity_.x = -velocity_.x;
		velocity_.y = -velocity_.y;
		velocity_.z = -velocity_.z;
	} else {
		if (mask & VEL_X)
			velocity_.x = -velocity_.x;
		if (mask & VEL_Y)
			velocity_.y = -velocity_.y;
		if (mask & VEL_Z)
			velocity_.z = -velocity_.z;
	}

    if (mask & COLOR) {
		color_.r = -color_.r;
		color_.g = -color_.g;
		color_.b = -color_.b;
	};

    if (mask & MASS)
		mass_ = -mass_;
    if (mask & SIZE)
		size_ = -size_;
    if (mask & SPIN)
		spin_ = -spin_;
    if (mask & ROTATION)
		rotation_ = -rotation_;
    if (mask & DENSITY)
		density_ = -density_;
    if (mask & SAT)
		sat_ = -sat_;
}


// ================================================================================
// Interpolate between two particules  
// --------------------------------------------------------------------------------
//  
// ================================================================================
void Particle::interpolate (const unsigned mask, const Particle &p1, const Particle &p2, const float factor)
{
	float k2 = factor;
	if (k2 > 1) k2 = 1;
	if (k2 < 0) k2 = 0;
	float k1 = 1-k2;

	position_.x = p1.position_.x*k1 + p2.position_.x*k2;
	position_.y = p1.position_.y*k1 + p2.position_.y*k2;
	position_.z = p1.position_.z*k1 + p2.position_.z*k2;
	velocity_.x = p1.velocity_.x*k1 + p2.velocity_.x*k2;
	velocity_.y = p1.velocity_.y*k1 + p2.velocity_.y*k2;
	velocity_.z = p1.velocity_.z*k1 + p2.velocity_.z*k2;
	color_.r    = p1.color_.r*k1    + p2.color_.r*k2;
	color_.g    = p1.color_.g*k1    + p2.color_.g*k2;
	color_.b    = p1.color_.b*k1    + p2.color_.b*k2;
	density_    = p1.density_*k1    + p2.density_*k2;
	mass_       = p1.mass_*k1       + p2.mass_*k2;
	size_       = p1.size_*k1       + p2.size_*k2;
	rotation_   = p1.rotation_*k1   + p2.rotation_*k2;
	sat_        = p1.sat_*k1        + p2.sat_*k2;
	lifespan_   = p1.lifespan_*k1   + p2.lifespan_*k2;
	age_        = 0;
	dead_       = false;
}


// ================================================================================
// Fit particle into a sphere
// --------------------------------------------------------------------------------
//  
// ================================================================================
void Particle::fit_sphere (const vector &center, const float radius) 
{
	fit_sphere (center.x, center.y, center.z, radius);
}
void Particle::fit_sphere (const float cx, const float cy, const float cz, const float radius)
{
	float theta = (rand()/float(RAND_MAX)) * 2 * M_PI;
	float phi   = (rand()/float(RAND_MAX)) * M_PI;
	position_.x = cx + radius*sin(theta)*sin(phi);
	position_.y = cy + radius*cos(theta);
	position_.z = cz + radius*sin(theta)*cos(phi);
}


// ================================================================================
// Fit particle into a cube
// --------------------------------------------------------------------------------
//  
// ================================================================================
void Particle::fit_cube (const vector &center, const float size) 
{
	fit_cube (center.x, center.y, center.z, size);
}
void Particle::fit_cube (const float cx, const float cy, const float cz, const float size)
{
   position_.x = cx + (rand()/float(RAND_MAX))*2*size - size;
   position_.y = cy + (rand()/float(RAND_MAX))*2*size - size;
   position_.z = cz + (rand()/float(RAND_MAX))*2*size - size;
}


// ================================================================================
// bullet (make particle going toward target)
// --------------------------------------------------------------------------------
//  
// ================================================================================
void Particle::bullet (const vector &target)
{
	bullet (target.x, target.y, target.z);
}
void Particle::bullet (const float tx, const float ty, const float tz)
{
	float velocity = sqrt (velocity_.x*velocity_.x +velocity_.y*velocity_.y +velocity_.z*velocity_.z);
	velocity_.x = velocity * (tx - position_.x);
    velocity_.y = velocity * (ty - position_.y);
    velocity_.z = velocity * (tz - position_.z);
}


// ================================================================================
// linear gravity
// --------------------------------------------------------------------------------
//  
// ================================================================================
void Particle::linear_gravity (const vector &direction, const float gravity, const float dt)
{
	linear_gravity (direction.x, direction.y, direction.z, gravity, dt);
}
void Particle::linear_gravity (const float dx, const float dy, const float dz, const float gravity, const float dt)
{
	float dlength = sqrt (dx*dx + dy*dy + dz*dz);
    if (dlength == 0)
		dlength = 0.0001;
    velocity_.x = velocity_.x + (dx / dlength) * gravity * dt;
    velocity_.y = velocity_.y + (dy / dlength) * gravity * dt;
    velocity_.z = velocity_.z + (dz / dlength) * gravity * dt;
}

// ================================================================================
// circular gravity
// --------------------------------------------------------------------------------
//  
// ================================================================================
void Particle::circular_gravity (const vector &point, const float gravity, const float dt)
{
	circular_gravity (point.x, point.y, point.z, gravity, dt);
}
void Particle::circular_gravity (const float px, const float py, const float pz, const float gravity, const float dt)
{
	float dirx = px - position_.x;
	float diry = py - position_.y;
	float dirz = pz - position_.z;
	float dlength = sqrt (dirx*dirx + diry*diry + dirz*dirz);
    if (dlength == 0)
		dlength = 0.0001f;
	velocity_.x += (dirx / dlength) * gravity * dt;
	velocity_.y += (diry / dlength) * gravity * dt;
	velocity_.z += (dirz / dlength) * gravity * dt;
}


// ================================================================================
// point gravity
// --------------------------------------------------------------------------------
//  
// ================================================================================
void Particle::point_gravity (const vector &point, const float radius, const float mass, const float dt)
{
	point_gravity (point.x, point.y, point.z, radius, mass, dt);
}
void Particle::point_gravity (const float px, const float py, const float pz,  const float radius, const float mass, const float dt)
{
	float dx = px - position_.x;
	float dy = py - position_.y;
	float dz = pz - position_.z;
	float dlength = sqrt (dx*dx + dy*dy + dz*dz);
    if (dlength < radius)
		return;
	float g = ((dt*mass)/1000.0f)/(dlength*dlength*dlength);
	velocity_.x += dx * g;
	velocity_.y += dy * g;
	velocity_.z += dz * g;
}

// ================================================================================
// friction
// --------------------------------------------------------------------------------
//  
// ================================================================================
void Particle::friction (const float viscosity, const float dt)
{
	float speed = sqrt (velocity_.x*velocity_.x + velocity_.y*velocity_.y + velocity_.z*velocity_.z);
	if (speed == 0)
		return;
	float k = 0;
	float dspeed = 18.0f * viscosity * speed * dt / density_; // k ought to be about 6 * pi
	if (dspeed < speed)
		k = (speed - dspeed) / speed;
	velocity_.x *= k;
	velocity_.y *= k;
	velocity_.z *= k;
}


// ================================================================================
// flow
// --------------------------------------------------------------------------------
//  
// ================================================================================
void Particle::flow (const vector &direction, const float viscosity, const float dt)
{
	flow (direction.x, direction.y, direction.z, viscosity, dt);
}
void Particle::flow (const float dx, const float dy, const float dz, const float viscosity, const float dt)
{
	vector dvelocity;
	dvelocity.x = velocity_.x - dx;
	dvelocity.y = velocity_.y - dy;
	dvelocity.z = velocity_.z - dz;

	float speed = sqrt(dvelocity.x*dvelocity.x + dvelocity.y*dvelocity.y + dvelocity.z*dvelocity.z);
	if (speed == 0)
		return;
	float dspeed = 18.0f * viscosity * speed * dt / density_; // k ought to be about 6 * pi  = 18
	float k = 0;
	if (dspeed < speed)
		k = (speed - dspeed) / speed;

	dvelocity.x *= k;
	dvelocity.y *= k;
	dvelocity.z *= k;

	velocity_.x = dvelocity.x + dx;
	velocity_.y = dvelocity.y + dy;
	velocity_.z = dvelocity.z + dz;
}


// ================================================================================
// fade
// --------------------------------------------------------------------------------
//  
// ================================================================================
float Particle::fade (void)
{
	if ((cloud_->fade_in_stop() > 1.0f) || (cloud_->fade_out_start() > 1.0f) ||
		(cloud_->fade_in_stop() < 0.0f) || (cloud_->fade_out_start() < 0.0f))
		return 1.0f;
	if (cloud_->fade_in_stop() > cloud_->fade_out_start())
		return 1.0f;

	// fade in
	if (age_ < lifespan_*cloud_->fade_in_stop())
		return age_ / (lifespan_*cloud_->fade_in_stop());

	// fade out
	if (age_ > lifespan_*cloud_->fade_out_start())
		return (1 - age_) / (1 - lifespan_*cloud_->fade_out_start());

	return 1.0f;
}


// ================================================================================
// billboard
// --------------------------------------------------------------------------------
//  
// ================================================================================
void Particle::billboard (const bool mode) 
{
	if (mode == ON) {
		float m[16];
		glGetFloatv (GL_MODELVIEW_MATRIX, m);
		for (int i=0; i<3; i++)
			for (int j=0; j<3; j++)
				if (i == j)
					m[i*4+j] = 1.0f;
				else
					m[i*4+j] = 0.0f;
		glPushMatrix();
		glLoadMatrixf (m);
	}
	else
		glPopMatrix();
}


// ================================================================================
// render_point
// --------------------------------------------------------------------------------
//  
// ================================================================================
void Particle::render (const float alpha)
{
	float a;
    if (cloud_->blend() == FX_OPAQUE)
		a = 1;
	else
		a = sat_ * alpha * fade ();

    glPushMatrix();

	if (cloud_->type() == FX_POINT) {
		glColor4f (color_.r*a, color_.g*a, color_.b*a, alpha);
		glTranslatef (position_.x, position_.y, position_.z);
		glPointSize (2);
		//glEnable(GL_POINT_SMOOTH);
		glBegin(GL_POINTS);
		{
			glVertex3f (0.0f, 0.0f, 0.0f);
		}
		glEnd();
	}
	else if (cloud_->type() == FX_QUAD) {
		glColor4f (color_.r*a, color_.g*a, color_.b*a, alpha);
		glTranslatef (position_.x, position_.y, position_.z);
		billboard (ON);
		glRotatef (rotation_, 0.0f, 0.0f, 1.0f);
		glBegin (GL_QUADS);
		{
			glVertex3f (+size_, -size_, 0.0f);
			glVertex3f (-size_, -size_, 0.0f);
			glVertex3f (-size_, +size_, 0.0f);
			glVertex3f (+size_, +size_, 0.0f);
		}
		glEnd();
		billboard (OFF);
	}
	else if (cloud_->type() == FX_TEX_QUAD) {
		glColor4f (color_.r*a, color_.g*a, color_.b*a, alpha);
		glTranslatef (position_.x, position_.y, position_.z);
		billboard (ON);
		glRotatef (rotation_, 0.0f, 0.0f, 1.0f);
		glBegin (GL_QUADS);
		{
			glTexCoord2f (cloud_->tile().max_u, cloud_->tile().min_v);
			glVertex3f (size_, -size_, 0.0f);
			glTexCoord2f (cloud_->tile().min_u, cloud_->tile().min_v);
			glVertex3f (-size_, -size_, 0.0f);
			glTexCoord2f (cloud_->tile().min_u, cloud_->tile().max_v);
			glVertex3f (-size_, size_, 0.0f);
			glTexCoord2f (cloud_->tile().max_u, cloud_->tile().max_v);
			glVertex3f (size_, size_, 0.0f);
		}
		glEnd ();
		billboard (OFF);
	}
	else if (cloud_->type() == FX_SPARK) {
		float m[16];
		vector v;
		float vl, angle, tail;
		glTranslatef (position_.x, position_.y, position_.z);
		glGetFloatv (GL_MODELVIEW_MATRIX, m);
		v.x = velocity_.x * m[0] + velocity_.y * m[4] + velocity_.z * m[8];
		v.y = velocity_.x * m[1] + velocity_.y * m[5] + velocity_.z * m[9];
		vl  = sqrt (v.x*v.x+v.y*v.y);
		tail = 1+vl*cloud_->elongation();
		angle = acos (v.y/vl)*360.0f / (2*M_PI);
		if (v.x < 0)
			angle = -angle;
		billboard(ON);
		glRotatef (-angle, 0.0f, 0.0f, 1.0f);
		glBegin (GL_QUADS);
		{
			glColor4f (0.0f, 0.0f, 0.0f, 0.0f);
			glVertex3f (size_, - size_*tail, 0.0f);
			glVertex3f (-size_, -size_*tail, 0.0f);
			glColor4f (color_.r*a, color_.g*a, color_.b*a, alpha);
			glVertex3f (-size_, size_, 0.0f);
			glVertex3f (size_, size_, 0.0f);
		}
		glEnd ();
		billboard (OFF);
	}
	else if (cloud_->type() == FX_TEX_SPARK) {
		float m[16];
		vector v;
		float vl, angle, tail;
		glTranslatef (position_.x, position_.y, position_.z);
		glGetFloatv (GL_MODELVIEW_MATRIX, m);
		v.x = velocity_.x * m[0] + velocity_.y * m[4] + velocity_.z * m[8];
		v.y = velocity_.x * m[1] + velocity_.y * m[5] + velocity_.z * m[9];
		vl  = sqrt (v.x*v.x+v.y*v.y);
		tail = 1+vl*cloud_->elongation();
		angle = acos (v.y/vl)*360.0f / (2*M_PI);
		if (v.x < 0)
			angle = -angle;
		billboard (ON);
		glRotatef (-angle, 0.0f, 0.0f, 1.0f);
		glBegin (GL_QUADS);
		{
			glColor4f (color_.r*a, color_.g*a, color_.b*a, alpha);
			glTexCoord2f (cloud_->tile().max_u, cloud_->tile().min_v);
			glVertex3f (size_, - size_*tail, 0.0f);
			glTexCoord2f (cloud_->tile().min_u, cloud_->tile().min_v);
			glVertex3f (-size_, -size_*tail, 0.0f);
			glTexCoord2f (cloud_->tile().min_u, cloud_->tile().max_v);
			glVertex3f (-size_, size_, 0.0f);
			glTexCoord2f (cloud_->tile().max_u, cloud_->tile().max_v);
			glVertex3f (size_, size_, 0.0f);
		}
		glEnd ();
		billboard (OFF);
	}
	else if (cloud_->type() == FX_SPRITE) {
		if (cloud_->blend() == FX_OPAQUE)
			glColor4f (color_.r, color_.g, color_.b, alpha);
		else
			glColor4f (color_.r, color_.g, color_.b, a);
		glTranslatef (position_.x, position_.y, position_.z);
		billboard (ON);
		glRotatef (-rotation_, 0.0f, 0.0f, 1.0f);
		glBegin(GL_QUADS);
		{
			glTexCoord2f (cloud_->tile().max_u, cloud_->tile().min_v);
			glVertex3f  (size_,-size_, 0.0f);
			glTexCoord2f (cloud_->tile().min_u, cloud_->tile().min_v);
			glVertex3f (-size_,-size_, 0.0f);
			glTexCoord2f (cloud_->tile().min_u, cloud_->tile().max_v);
			glVertex3f (-size_, size_, 0.0f);
			glTexCoord2f (cloud_->tile().max_u, cloud_->tile().max_v);
			glVertex3f (size_, size_, 0.0f);
		}
		glEnd ();
		billboard (OFF);
	}
	glPopMatrix();
}
