/* 
 * Copyright (c) 2005 Nicolas Rougier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this software; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <string>
#include <sstream>
#include <iomanip>
#include <cstring>
#include <cstdlib>
#include <SDL/SDL.h>
#include <SDL/SDL_opengl.h>
#include <png.h>

#include "camera.h"
#include "texture.h"
#include "screen.h"



Screen::Screen (int width, int height, Uint32 bpp)
{
    _width      = width;
    _height     = height;
    _bpp        = bpp;
    _camera		= 0;    

    // SDL initialization
    if (SDL_Init (SDL_INIT_VIDEO) < 0) {
	    fprintf (stderr, "Video initialization failed: %s", SDL_GetError());
		SDL_Quit();
		exit (EXIT_FAILURE);
	}
    // Collect information about video
    const SDL_VideoInfo *info = SDL_GetVideoInfo(); 
    if (!info) {
	    fprintf (stderr, "Video query failed: %s", SDL_GetError());
		SDL_Quit();
		exit (EXIT_FAILURE);
	}

    _flags  = SDL_OPENGL;
    _flags |= SDL_GL_DOUBLEBUFFER;
    _flags |= SDL_HWPALETTE;
    _flags |= SDL_RESIZABLE;
    //    _flags |= SDL_FULLSCREEN;

    if (info->hw_available)
		_flags |= SDL_HWSURFACE;
    else
		_flags |= SDL_SWSURFACE;
    if (info->blit_hw)
		_flags |= SDL_HWACCEL;

    SDL_GL_SetAttribute (SDL_GL_DOUBLEBUFFER, 1);
    SDL_GL_SetAttribute (SDL_GL_RED_SIZE,     5);
    SDL_GL_SetAttribute (SDL_GL_GREEN_SIZE,   5);
    SDL_GL_SetAttribute (SDL_GL_BLUE_SIZE,    5);
    SDL_GL_SetAttribute (SDL_GL_DEPTH_SIZE,   16);


    // Try to get the best (biggest) resolution for fullscreen mode
    SDL_Rect **modes = SDL_ListModes (NULL, _flags);
    if (modes == (SDL_Rect **)0) {
        printf("No suitable modes available!\n");
        exit (EXIT_FAILURE);
    }
    if (modes == (SDL_Rect **)-1) {
        _fullwidth  = 1280;
        _fullheight = 1024;
    } else {
        _fullwidth  = modes[0]->w;
        _fullheight = modes[0]->h;
    }
    full();
    //resize (_width, _height);
}

Screen::~Screen()
{}

void
Screen::resize (int width, int height)
{
    _width  = width;
    _height = height;

    // Full screen mode
    if (_flags & SDL_FULLSCREEN) {
        _surface = SDL_SetVideoMode (_fullwidth, _fullheight, _bpp, _flags);
        
    } else {
        _surface = SDL_SetVideoMode (_width, _height, _bpp, _flags);
    }
    
    if (!_surface) {
        fprintf (stderr, "Video mode set failed: %s", SDL_GetError());
        SDL_Quit();
        exit (EXIT_FAILURE);
    }

    if (_camera)
        _camera->resize ();
    
#ifdef WIN32
    TextureManager::it()->init();
#endif
}

void
 Screen::update (void)
{
	SDL_GL_SwapBuffers ();
}

void
Screen::full (void)
{
    static int saved_width  = 0;
    static int saved_height = 0;

	_flags ^= SDL_FULLSCREEN;

    if (_flags & SDL_FULLSCREEN) {
        saved_width  = _width;
        saved_height = _height;

        float w = float (_fullwidth)  / float(_width);
        float h = float (_fullheight) / float(_height);
        if (w < h) {
            _width  = int(_width*w);
            _height = int(_height*w);
            
        } else {
            _width  = int(_width*h);
            _height = int(_height*h);
        }
    } else {
        if (saved_width)
            _width  = saved_width;
        if (saved_height)
            _height = saved_height;
    }

    resize (_width, _height);
}

void
Screen::blit_surface (SDL_Surface *surface, SDL_Rect &src, SDL_Rect &dst, float alpha)
{
	TextureManager *tmanager = TextureManager::it();

	glEnable (GL_TEXTURE_2D);

	// Surface with no alpha channel
	if (surface->format->BytesPerPixel == 3) {
		glBindTexture (GL_TEXTURE_2D, tmanager->get("RGB_surface")->get_id());
		glTexSubImage2D (GL_TEXTURE_2D, 0,
						 0, 0, surface->w, surface->h,
						 GL_RGB, GL_UNSIGNED_BYTE, surface->pixels);
	}
	// Surface with alpha channel
	else if (surface->format->BytesPerPixel == 4) {
		glBindTexture (GL_TEXTURE_2D, tmanager->get("RGBA_surface")->get_id());
		glTexSubImage2D (GL_TEXTURE_2D, 0,
						 0, 0, surface->w, surface->h,
						 GL_RGBA, GL_UNSIGNED_BYTE, surface->pixels);
	}
	// Not handled
	else {
		return;
	}

	glDisable(GL_DEPTH_TEST);							// Disables Depth Testing
	glEnable (GL_BLEND);								// Enable blend
    glColor4f(1.0f,1.0f,1.0f,alpha);
    glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);	// Blend function

	glMatrixMode (GL_PROJECTION);						// Select The Projection Matrix

    

	glPushMatrix ();									// Store The Projection Matrix
	{
		glLoadIdentity ();								// Reset The Projection Matrix
		glOrtho (0, _width, 0, _height, -1, 1);			// Set Up An Ortho Screen
		glMatrixMode (GL_MODELVIEW);					// Select The Modelview Matrix
		glPushMatrix ();								// Store The Modelview Matrix
		{
			glLoadIdentity ();							// Reset The Modelview Matrix
			glTranslated (dst.x, _height-dst.y-src.h, 0);

			glBegin (GL_QUADS);
			{
				// Texture Coord (Top Left)
				glTexCoord2f (float(src.x)/1024.0f, float(src.y+src.h)/1024.0f);
				// Vertex Coord (Bottom Left)
				glVertex2i (0, 0);

				// Texture Coord (Top Right)
				glTexCoord2f (float(src.x+src.w)/1024.0f, float(src.y+src.h)/1024.0f);
				// Vertex Coord (Bottom Right)
				glVertex2i (dst.w, 0);

				// Texture Coord (Bottom Right)
				glTexCoord2f (float(src.x+src.w)/1024.0f, float(src.y)/1024.0f);
				// Vertex Coord (Top Right)
				glVertex2i (dst.w, dst.h);
				
				// Texture Coord (Bottom Left)
				glTexCoord2f (float(src.x)/1024.0f, float(src.y)/1024.0f);
				// Vertex Coord (Top Left)	
				glVertex2i (0, dst.h);
			}
			glEnd ();
			glMatrixMode (GL_PROJECTION);				// Select The Projection Matrix
		}
		glPopMatrix ();									// Restore The Old Projection Matrix
		glMatrixMode (GL_MODELVIEW);					// Select The Modelview Matrix
	}
	glPopMatrix();										// Restore The Old Projection Matrix
	glDisable (GL_BLEND);								// Disable blend
	glEnable (GL_DEPTH_TEST);							// Enables Depth Testing
}

int
Screen::screenshot (void)
{
    char filename[256];
    for (int i=0; i<100; i++) {
        sprintf (filename, "screenshot-%.2d.png", i);
        FILE *file = fopen (filename, "r");
        if (file == NULL)
            break;
    }

    if (!(_flags & SDL_OPENGL)) {
        export_png (_surface, filename);
    } else {
        SDL_Surface *tmp;
        unsigned char *pixels = 0;
        tmp = SDL_CreateRGBSurface (SDL_SWSURFACE, _surface->w, _surface->h, 24,
#if SDL_BYTEORDER == SDL_LIL_ENDIAN
		0x000000FF, 0x0000FF00, 0x00FF0000, 0
#else
		0x00FF0000, 0x0000FF00, 0x000000FF, 0
#endif
                                    );
        if (tmp == NULL)
            return -1;
        pixels = new unsigned char[3*_surface->w*_surface->h];
        if (!pixels) {
            SDL_FreeSurface (tmp);
            return -1;
        }
        glReadPixels (0, 0, _surface->w, _surface->h, GL_RGB, GL_UNSIGNED_BYTE, pixels);

        for (int i=0; i<_height; i++)
            memcpy (((char *) tmp->pixels) + tmp->pitch * i, pixels + 3*_surface->w * ( _surface->h-i-1), _surface->w*3);

		delete [] pixels;
        export_png (tmp, filename);
        SDL_FreeSurface (tmp);
    }
    return 0;
}

// ================================================================================
//  
// --------------------------------------------------------------------------------
//  
// ================================================================================
int
Screen::export_png (SDL_Surface *src, std::string filename)
{
    SDL_Surface *srcRGB =  SDL_CreateRGBSurface (SDL_SWSURFACE, src->w, src->h, 24,
                                                 0xFF000000, 0x0000FF00, 0x00FF0000, 0x000000FF);
    SDL_BlitSurface (src, 0, srcRGB, 0);
    unsigned char *pixels = (unsigned char *) srcRGB->pixels;

    png_bytep *row_pointers = new png_bytep[srcRGB->h];
	for (int i=0; i<srcRGB->h; i++)
        row_pointers[i] = (png_bytep) (pixels + i*srcRGB->pitch);

	FILE *fp = fopen (filename.c_str(), "wb");
	png_structp png_ptr = png_create_write_struct (PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
	if (!png_ptr)
		return 0;
	png_infop info_ptr = png_create_info_struct (png_ptr);
    if (!info_ptr)
		return 0;
	if (setjmp (png_jmpbuf (png_ptr)))
        return 0;
	png_init_io(png_ptr, fp);
    if (setjmp (png_jmpbuf (png_ptr)))
		return 0;

    png_set_compression_level (png_ptr, Z_BEST_SPEED);
    //    png_set_compression_level (png_ptr, Z_BEST_COMPRESSION);
	png_set_IHDR (png_ptr, info_ptr, srcRGB->w, srcRGB->h,
                  8, PNG_COLOR_TYPE_RGB, PNG_INTERLACE_NONE,
                  PNG_COMPRESSION_TYPE_BASE, PNG_FILTER_TYPE_BASE);
    png_write_info (png_ptr, info_ptr);

	if (setjmp (png_jmpbuf (png_ptr)))
		return 0;
	png_write_image (png_ptr, row_pointers);
	if (setjmp (png_jmpbuf (png_ptr)))
        return 0;
	png_write_end(png_ptr, NULL);
    fclose(fp);

    delete [] row_pointers;
    SDL_FreeSurface (srcRGB);
    
    return 0;
}
