/* 
 * Copyright (c) 2004 Nicolas Rougier
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this software; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <cmath>
#include <SDL/SDL_image.h>
#include "log.h"
#include "texture.h"



Texture::Texture (std::string filename, bool alpha, bool mipmaps, bool repeat)
{
    _filename = filename;
    _alpha    = alpha;
    _mipmaps  = mipmaps;
    _repeat   = repeat;
    _id       = 0;
}

void
Texture::generate (void)
{
	GLubyte *data = 0;
	int width, height;

	if ((data = load (width, height)) != 0) {

		int type = (_alpha) ? GL_RGBA : GL_RGB;
		glGenTextures (1, &_id);
		glBindTexture (GL_TEXTURE_2D, _id); 

		glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);   // texture blends with object background
		//  glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL);  // texture does NOT blend with object background
		glTexImage2D (GL_TEXTURE_2D, 0, type, width, height, 0, type, GL_UNSIGNED_BYTE, data);
		int filter_min, filter_mag;
		filter_min = (_mipmaps) ? GL_NEAREST_MIPMAP_NEAREST : GL_NEAREST;
		filter_mag = GL_NEAREST;

		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, filter_min);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, filter_mag);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, (_repeat) ? GL_REPEAT : GL_CLAMP);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, (_repeat) ? GL_REPEAT : GL_CLAMP);
		if (_mipmaps)
			gluBuild2DMipmaps(GL_TEXTURE_2D, type, width, height, type, GL_UNSIGNED_BYTE, data);
		delete [] data;
	}
}


// ================================================================================
//  load
// --------------------------------------------------------------------------------
//  Returned pointer on byte array must be freed by caller
// ================================================================================
GLubyte *
Texture::load (int &width, int &height)
{
	SDL_Surface *img = IMG_Load (_filename.c_str());
	if(!img) {
		fprintf(stderr, "Unable to load texture from %s!\n%s\n", _filename.c_str(), IMG_GetError());
		width = 0;
		height = 0;
		return 0;
	}
	width = img->w;
	height = img->h;

	// Build image data from the surface (RGB == 3 bytes, RGBA == 4 bytes)
	Uint32 dim = img->w * img->h * ((_alpha) ? 4 : 3);
	GLubyte *data = new GLubyte[dim];
	if(!data) {
		fprintf(stderr, "Unable to create a texture from %s\n", _filename.c_str());
		width = 0;
		height = 0;
		return 0;
	}

	// Traverse trough surface and grab the pixels
	int pos = 0;
	for (int y=(img->h-1); y>-1; y--) {
		for (int x=0; x<img->w; x++) {
			Uint8 r, g, b, a;
			Uint32 color = pixel (img, x, y);
			if (!_alpha)
				SDL_GetRGB (color, img->format, &r, &g, &b);
			else
				SDL_GetRGBA (color, img->format, &r, &g, &b, &a);
			data[pos] = r; pos++;
			data[pos] = g; pos++;
			data[pos] = b; pos++;
			if (_alpha) {
				data[pos] = a; pos++;
			}
		}
	}

	SDL_FreeSurface (img);
	return data;
}



// ================================================================================
//  pixel
// --------------------------------------------------------------------------------
//  @param surface is the surface to get pixel from
//  @param x is the x coordinate of the pixel to get
//  @param y is the y coordinate of the pixel to get
//  @return the pixel value at (x,y) in surface
// ================================================================================
Uint32
Texture::pixel (SDL_Surface *surface, int x, int y)
{
	int bpp = surface->format->BytesPerPixel;
	Uint8 *p = (Uint8 *) surface->pixels + y*surface->pitch + x*bpp;

	switch (bpp) {
	case 1:
		return *p;
	case 2:
		return *(Uint16*)p;
	case 3:
		if(SDL_BYTEORDER == SDL_BIG_ENDIAN)
			return p[0] << 16 | p[1] << 8 | p[2];
		else
			return p[0] | p[1] << 8 | p[2] << 16;
	case 4:
		return *(Uint32*)p;
	default:
		return 0;
	}
}



// ================================================================================
//  Initialization
// --------------------------------------------------------------------------------
//  We create a default texture that will be used when a load fails.
// ================================================================================
void
TextureManager::init (void)
{
    Texture *tex_default = 0;
    Texture *tex_RGB     = 0;
    Texture *tex_RGBA    = 0;
    
    for (unsigned int i=0; i<database_.size(); i++) {
        if (database_[i]->get_filename() == "default")
            tex_default = database_[i];
        else if (database_[i]->get_filename() == "RGB_surface")
            tex_RGB = database_[i];
        else if (database_[i]->get_filename() == "RGBA_surface")
            tex_RGBA = database_[i];
        else
            database_[i]->generate();
    }

    if (!tex_default) {
        tex_default = new Texture ("default");
        database_.push_back (tex_default);
    }
    if (!tex_RGB) {
        tex_RGB = new Texture ("RGB_surface", false);
        database_.push_back (tex_RGB);
    }
    if (!tex_RGBA) {
        tex_RGBA = new Texture ("RGBA_surface", true);
        database_.push_back (tex_RGBA);
    }

    
	// Default texture entry
	GLuint id;
	glGenTextures (1, &id);
	glBindTexture (GL_TEXTURE_2D, id);
	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
    tex_default->get_id() = id;

	// Make the default texture image (128x128, RGBA)
	GLubyte *checker = new GLubyte [128*128*4];
	for (GLuint i = 0; i<128; i++) {
		for (GLuint j = 0; j < 128; j++) {
			int c = ( !(i & 8) ^ !(j & 8)) * 255;	  
			checker [(i*128*4) + (j*4) + 0] = (GLubyte)c;
			checker [(i*128*4) + (j*4) + 1] = (GLubyte)c;
			checker [(i*128*4) + (j*4) + 2] = (GLubyte)c;
			checker [(i*128*4) + (j*4) + 3] = (GLubyte)255;
		}
	}
	glTexImage2D (GL_TEXTURE_2D, 0, 4, 128, 128, 0, GL_RGBA, GL_UNSIGNED_BYTE, checker);
	delete [] checker;


	// Default texture entry for blit surface screen function (no alpha)
	glGenTextures (1, &id);
	glBindTexture (GL_TEXTURE_2D, id);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
	GLubyte *data = new GLubyte [1024*1024*3];
	glTexImage2D (GL_TEXTURE_2D, 0, 3, 1024, 1024, 0, GL_RGB, GL_UNSIGNED_BYTE, data);
    tex_RGB->get_id() = id;
	delete [] data;


	// Default texture entry for blit surface screen function (no alpha)
	glGenTextures (1, &id);
	glBindTexture (GL_TEXTURE_2D, id);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
	data = new GLubyte [1024*1024*4];
	glTexImage2D (GL_TEXTURE_2D, 0, 4, 1024, 1024, 0, GL_RGBA, GL_UNSIGNED_BYTE, data);
    tex_RGBA->get_id() = id;
	delete [] data;
}


// ================================================================================
//  get
// --------------------------------------------------------------------------------
//  Try to get the texture from database and if not available, try to load it.
//  If load fails, return the default checker texture.
// ================================================================================
Texture *
TextureManager::get (std::string filename, bool alpha, bool mipmaps, bool repeat)
{
     for (unsigned int i=0; i<database_.size(); i++)
         if ((database_[i]->get_filename() == filename))
             return database_[i];
             
	// Get rid of '\' (dos/windows style)
	for (unsigned int i=0; i<filename.size(); i++)
		if (filename[i] == '\\')
			filename[i] = '/';

    Texture *tex = new Texture (filename, alpha, mipmaps, repeat);
    database_.push_back (tex);
    tex->generate();
    return tex;
}
